@@ -32,6 +32,7 @@
 #include "file-data.h"
 #include "file-list.h"
 #include "file-utils.h"
+#include "glib-utils.h"
 #include "fr-archive.h"
 #include "fr-command.h"
 #include "fr-command-ar.h"
@@ -51,12 +52,15 @@
 #include "fr-process.h"
 #include "utf8-fnmatch.h"
 
+#ifndef NCARGS
+#define NCARGS _POSIX_ARG_MAX
+#endif
 
 #define g_signal_handlers_disconnect_by_data(instance, data) \
     g_signal_handlers_disconnect_matched ((instance), G_SIGNAL_MATCH_DATA, \
                                           0, 0, NULL, NULL, (data))
 
-#define MAX_CHUNK_LEN (NCARGS - PATH_MAX) /* Max length of the command line */
+#define MAX_CHUNK_LEN (NCARGS * 2 / 3) /* Max command line length */
 #define UNKNOWN_TYPE "application/octet-stream"
 #define SAME_FS (FALSE)
 #define NO_BACKUP_FILES (FALSE)
@@ -182,11 +186,21 @@ fr_archive_stoppable (FRArchive *archive,
 }
 
 
+static gboolean
+fr_archive_add_is_stoppable (FRArchive *archive)
+{
+	if (archive->add_is_stoppable_func != NULL)
+		return (*archive->add_is_stoppable_func) (archive, archive->add_is_stoppable_data);
+	else
+		return FALSE;
+}
+
+
 static gboolean
 archive_sticky_only_cb (FRProcess *process,
 			FRArchive *archive)
 {
-	fr_archive_stoppable (archive, FALSE);
+	fr_archive_stoppable (archive, fr_archive_add_is_stoppable (archive));
 	return TRUE;
 }
 
@@ -202,6 +216,9 @@ fr_archive_init (FRArchive *archive)
 	archive->fake_load_func = NULL;
 	archive->fake_load_data = NULL;
 
+	archive->add_is_stoppable_func = NULL;
+	archive->add_is_stoppable_data = NULL;
+
 	archive->process = fr_process_new ();
 	g_signal_connect (G_OBJECT (archive->process), 
 			  "sticky_only",
@@ -313,9 +330,9 @@ get_mime_type_from_content (const char *filename)
 
 
 static gboolean
-hexcmp (const guchar *first_bytes,
-	const guchar *buffer,
-	int           len)
+hexcmp (const char *first_bytes,
+	const char *buffer,
+	int         len)
 {
 	int i;
 	
@@ -338,8 +355,8 @@ get_mime_type_from_sniffer (const char *filename)
 	}            sniffer_data [] = { 
 		{"application/zip",                   "\x50\x4B\x03\x04", 4},
 		/* FIXME
-		 *{"application/x-compressed-tar",      "\x1F\x8B\x08\x08", 4},
-		 *{"application/x-bzip-compressed-tar", "\x42\x5A\x68\x39", 4},
+		   {"application/x-compressed-tar",      "\x1F\x8B\x08\x08", 4},
+		   {"application/x-bzip-compressed-tar", "\x42\x5A\x68\x39", 4},
 		 */
 		{ NULL, NULL, 0 } 
 	};
@@ -536,7 +553,7 @@ action_started (FRCommand *command,
 		FRAction   action,
 		FRArchive *archive)
 {
-	g_print ("FRArchive::action_started: %d\n", action);
+	debug (DEBUG_INFO, "FRArchive::action_started: %d\n", action);
 	g_signal_emit (G_OBJECT (archive), 
 		       fr_archive_signals[START],
 		       0,
@@ -572,8 +589,11 @@ action_performed (FRCommand   *command,
 	case FR_ACTION_GET_LIST:
 		s_action = "Get list";
 		break;
+	default:
+		s_action = "";
+		break;
 	}
-	g_print ("%s [DONE] (FR::Archive)\n", s_action);
+	debug (DEBUG_INFO, "%s [DONE] (FR::Archive)\n", s_action);
 #endif
 
 	g_signal_emit (G_OBJECT (archive), 
@@ -706,6 +726,7 @@ fr_archive_load (FRArchive   *archive,
 	/* prefer mime-magic */
 
 	mime_type = get_mime_type_from_sniffer (filename);
+
 	if (mime_type == NULL)
 		mime_type = get_mime_type_from_content (filename);
 	
@@ -726,6 +747,10 @@ fr_archive_load (FRArchive   *archive,
 		g_object_unref (tmp_command);
 	}
 
+	if ((archive->command->file_type == FR_FILE_TYPE_ZIP)
+	    && (!is_program_in_path ("zip"))) 
+		archive->read_only = TRUE;
+
 	g_signal_connect (G_OBJECT (archive->command), 
 			  "start",
 			  G_CALLBACK (action_started),
@@ -813,10 +838,8 @@ create_tmp_base_dir (const char *base_dir,
 	if (dest_dir[strlen (dest_dir) - 1] == G_DIR_SEPARATOR)
 		dest_dir[strlen (dest_dir) - 1] = 0;
 
-#ifdef DEBUG
-	g_print ("base_dir: %s\n", base_dir);
-	g_print ("dest_dir: %s\n", dest_dir);
-#endif
+	debug (DEBUG_INFO, "base_dir: %s\n", base_dir);
+	debug (DEBUG_INFO, "dest_dir: %s\n", dest_dir);
 
 	temp_dir = get_temp_work_dir ();
 	tmp = remove_level_from_path (dest_dir);
@@ -825,18 +848,14 @@ create_tmp_base_dir (const char *base_dir,
 
 	ensure_dir_exists (parent_dir, 0700);
 
-#ifdef DEBUG
-	g_print ("mkdir %s\n", parent_dir);
-#endif
+	debug (DEBUG_INFO, "mkdir %s\n", parent_dir);
 
 	g_free (parent_dir);
 
 	dir = g_build_filename (temp_dir, "/", dest_dir, NULL);
 	symlink (base_dir, dir);
 
-#ifdef DEBUG
-	g_print ("symlink %s --> %s\n", dir, base_dir);
-#endif
+	debug (DEBUG_INFO, "symlink %s --> %s\n", dir, base_dir);
 
 	g_free (dir);
 	g_free (dest_dir);
@@ -938,6 +957,16 @@ newer_files_only (FRArchive  *archive,
 }
 
 
+void
+fr_archive_set_add_is_stoppable_func (FRArchive     *archive,
+				      FakeLoadFunc   func,
+				      gpointer       data)
+{
+	archive->add_is_stoppable_func = func;
+	archive->add_is_stoppable_data = data;
+}
+
+
 /* Note: all paths unescaped. */
 void
 fr_archive_add (FRArchive     *archive, 
@@ -981,7 +1010,7 @@ fr_archive_add (FRArchive     *archive,
 	} else
 		new_file_list = file_list;
 
-	fr_archive_stoppable (archive, FALSE);
+	fr_archive_stoppable (archive, fr_archive_add_is_stoppable (archive));
 
 	/* if the command cannot update,  get the list of files that are 
 	 * newer than the ones in the archive. */
@@ -995,9 +1024,7 @@ fr_archive_add (FRArchive     *archive,
 	} 
 
 	if (new_file_list == NULL) {
-#ifdef DEBUG
-		g_print ("nothing to update.\n");
-#endif
+		debug (DEBUG_INFO, "nothing to update.\n");
 
 		if (base_dir_created) 
 			rmdir_recursive (tmp_base_dir);
@@ -1215,7 +1242,7 @@ fr_archive_add_with_wildcard (FRArchive     *archive,
 
 typedef struct {
 	FRArchive     *archive;
-	char          *directory;
+	GList         *dir_list;
 	char          *base_dir;
 	char          *dest_dir;
 	gboolean       update;
@@ -1247,7 +1274,7 @@ add_directory__step2 (GList *file_list, gpointer data)
 
 	/**/
 
-	g_free (ad_data->directory);
+	path_list_free (ad_data->dir_list);
 	g_free (ad_data->base_dir);
 	g_free (ad_data->password);
 	g_free (ad_data);
@@ -1276,7 +1303,7 @@ fr_archive_add_directory (FRArchive     *archive,
 
 	ad_data = g_new0 (AddDirectoryData, 1);
 	ad_data->archive = archive;
-	ad_data->directory = g_strdup (directory);
+	ad_data->dir_list = g_list_prepend (NULL, g_strdup (directory));
 	ad_data->base_dir = g_strdup (base_dir);
 	ad_data->dest_dir = g_strdup (dest_dir);
 	ad_data->update = update;
@@ -1285,10 +1312,48 @@ fr_archive_add_directory (FRArchive     *archive,
 	ad_data->done_func = done_func;
 	ad_data->done_data = done_data;
 
-	return get_directory_file_list_async (directory, 
-					      base_dir, 
-					      add_directory__step2, 
-					      ad_data);
+	return get_items_file_list_async (ad_data->dir_list,
+					  base_dir, 
+					  add_directory__step2, 
+					  ad_data);
+}
+
+
+/* Note: all paths unescaped. */
+VisitDirHandle *
+fr_archive_add_items (FRArchive     *archive, 
+		      GList         *item_list,
+		      const char    *base_dir,
+		      const char    *dest_dir,
+		      gboolean       update,
+		      const char    *password,
+		      FRCompression  compression,
+		      DoneFunc       done_func,
+		      gpointer       done_data)
+	
+{
+	AddDirectoryData *ad_data;
+
+	if (archive->read_only)
+		return NULL;
+
+	fr_archive_stoppable (archive, TRUE);
+
+	ad_data = g_new0 (AddDirectoryData, 1);
+	ad_data->archive = archive;
+	ad_data->dir_list = path_list_dup (item_list);
+	ad_data->base_dir = g_strdup (base_dir);
+	ad_data->dest_dir = g_strdup (dest_dir);
+	ad_data->update = update;
+	ad_data->password = g_strdup (password);
+	ad_data->compression = compression;
+	ad_data->done_func = done_func;
+	ad_data->done_data = done_data;
+
+	return get_items_file_list_async (ad_data->dir_list, 
+					  base_dir, 
+					  add_directory__step2, 
+					  ad_data);
 }
 
 
@@ -1511,9 +1576,7 @@ compute_base_path (const char *base_dir,
 			new_path = g_strdup (file_name_from_path (path));
 		else
 			new_path = g_strdup (path);
-#ifdef DEBUG
-		g_print ("%s, %s --> %s\n", base_dir, path, new_path);
-#endif
+		debug (DEBUG_INFO, "%s, %s --> %s\n", base_dir, path, new_path);
 		return new_path;
 	}
 
@@ -1532,9 +1595,7 @@ compute_base_path (const char *base_dir,
 		new_path = g_strndup (path, name_len);
 	}
 
-#ifdef DEBUG
-	g_print ("%s, %s --> %s\n", base_dir, path, new_path);
-#endif
+	debug (DEBUG_INFO, "%s, %s --> %s\n", base_dir, path, new_path);
 
 	return new_path;
 }
@@ -1617,6 +1678,7 @@ fr_archive_extract (FRArchive  *archive,
 	if (extract_all && ! archive->command->propCanExtractAll) {
 		GList *scan;
 
+		file_list = NULL;
 		scan = archive->command->file_list;
 		for (; scan; scan = scan->next) {
 			FileData *fdata = scan->data;
@@ -1661,6 +1723,7 @@ fr_archive_extract (FRArchive  *archive,
 	if (extract_all && ! file_list_created) {
 		GList *scan;
 
+		file_list = NULL;
 		scan = archive->command->file_list;
 		for (; scan; scan = scan->next) {
 			FileData *fdata = scan->data;
@@ -1695,9 +1758,7 @@ fr_archive_extract (FRArchive  *archive,
 		else
 			sprintf (dest_filename, "%s/%s", dest_dir, filename);
 		
-#ifdef DEBUG		
-		g_print ("-> %s\n", dest_filename);
-#endif
+		debug (DEBUG_INFO, "-> %s\n", dest_filename);
 
 		/**/
 		
@@ -1717,11 +1778,7 @@ fr_archive_extract (FRArchive  *archive,
 
 	if (filtered == NULL) {
 		/* all files got filtered, do nothing. */
-
-#ifdef DEBUG
-		g_print ("All files got filtered, do nothing.\n");
-#endif
-
+		debug (DEBUG_INFO, "All files got filtered, do nothing.\n");
 		if (extract_all) 
 			path_list_free (file_list);
 		return;
