@@ -32,8 +32,9 @@
 #include <glib.h>
 #include "fr-process.h"
 #include "fr-marshal.h"
+#include "glib-utils.h"
 
-#define REFRESH_RATE 10
+#define REFRESH_RATE 20
 
 enum {
 	START,
@@ -385,6 +386,24 @@ fr_process_set_end_func (FRProcess    *fr_proc,
 }
 
 
+void
+fr_process_set_continue_func (FRProcess    *fr_proc, 
+			      ContinueFunc  func,
+			      gpointer      func_data)
+{
+	FRCommandInfo *c_info;
+
+	g_return_if_fail (fr_proc != NULL);
+
+	if (fr_proc->current_comm < 0)
+		return;
+
+	c_info = g_ptr_array_index (fr_proc->comm, fr_proc->current_comm);
+	c_info->continue_func = func;
+	c_info->continue_data = func_data;
+}
+
+
 void
 fr_process_end_command (FRProcess *fr_proc)
 {
@@ -566,20 +585,15 @@ start_current_command (FRProcess *fr_proc)
 	char          **argv;
 	int             i = 0;
 
-#ifdef DEBUG
-	g_print ("%d/%d) ", fr_proc->current_command, fr_proc->n_comm);
-#endif
+	debug (DEBUG_INFO, "%d/%d) ", fr_proc->current_command, fr_proc->n_comm);
 	
 	c_info = g_ptr_array_index (fr_proc->comm, 
 				    fr_proc->current_command);
 	arg_list = c_info->args;
 	dir = c_info->dir;
 
-	if (dir != NULL) {
-#ifdef DEBUG
-		g_print ("cd %s\n", dir); 
-#endif
-	}
+	if (dir != NULL) 
+		debug (DEBUG_INFO, "cd %s\n", dir); 
 	
 	command = NULL;
 	
@@ -689,6 +703,7 @@ check_child (gpointer data)
 	FRCommandInfo  *c_info;
 	pid_t           pid;
 	int             status;
+	gboolean        continue_process;
 
 	c_info = g_ptr_array_index (fr_proc->comm, fr_proc->current_command);
 
@@ -711,10 +726,9 @@ check_child (gpointer data)
 
 	if (c_info->ignore_error) {
 		fr_proc->error.type = FR_PROC_ERROR_NONE;
-		g_print ("[ignore error]\n");
-	}
+		debug (DEBUG_INFO, "[ignore error]\n");
 
-	if (fr_proc->error.type != FR_PROC_ERROR_STOPPED) {
+	} else if (fr_proc->error.type != FR_PROC_ERROR_STOPPED) {
 		if (WIFEXITED (status)) {
 			if (WEXITSTATUS (status) == 0)
 				fr_proc->error.type = FR_PROC_ERROR_NONE;
@@ -746,23 +760,30 @@ check_child (gpointer data)
 	if (c_info->end_func != NULL)
 		(*c_info->end_func) (c_info->end_data);
 
+	/* Check whether to continue or stop the process */
+
+	continue_process = TRUE;
+	if (c_info->continue_func != NULL)
+		continue_process = (*c_info->continue_func) (c_info->continue_data);
+
 	/* Execute next command. */
-
-	if (fr_proc->error.type != FR_PROC_ERROR_NONE) 
-		allow_sticky_processes_only (fr_proc, TRUE);
-
-	if (fr_proc->sticky_only) {
-		do {
+	if (continue_process) {
+		if (fr_proc->error.type != FR_PROC_ERROR_NONE) 
+			allow_sticky_processes_only (fr_proc, TRUE);
+		
+		if (fr_proc->sticky_only) {
+			do {
+				fr_proc->current_command++;
+			} while ((fr_proc->current_command <= fr_proc->n_comm)
+				 && ! command_is_sticky (fr_proc, 
+							 fr_proc->current_command));
+		} else
 			fr_proc->current_command++;
-		} while ((fr_proc->current_command <= fr_proc->n_comm)
-			 && ! command_is_sticky (fr_proc, 
-						 fr_proc->current_command));
-	} else
-		fr_proc->current_command++;
 
-	if (fr_proc->current_command <= fr_proc->n_comm) {
-		start_current_command (fr_proc);
-		return FALSE;
+		if (fr_proc->current_command <= fr_proc->n_comm) {
+			start_current_command (fr_proc);
+			return FALSE;
+		}
 	}
 
 	/* Done */
